import torch

from dlhub_sdk.models.servables.python import BasePythonServableModel
from dlhub_sdk.utils.types import compose_argument_block


class TorchModel(BasePythonServableModel):
    """Servable based on a Torch Model object.

    Assumes that the model has been saved to a pt or a pth file"""

    @classmethod
    def create_model(cls, model_path, input_shape, output_shape,
                     input_type='float', output_type='float'):
        """Initialize a PyTorch model.

        Args:
            model_path (string): Path to the pt or pth file that contains the weights and
                the architecture
            input_shape (tuple or [tuple]): Shape of input matrix to model
            output_shape (tuple or [tuple]): Shape of output matrix from model
            input_type (str or [str]): Data type of inputs
            output_type (str or [str]): Data type of outputs
       """
        output = super(TorchModel, cls).create_model('__call__')

        # Add model as a file to be sent
        output.add_file(model_path, 'model')

        # Get the model details
        if model_path.endswith('.pt') or model_path.endswith('.pth'):
            model = torch.load(model_path)
        else:
            raise ValueError('File type for architecture not recognized')

        # Get the inputs of the model
        output['servable']['methods']['run']['input'] =\
            output.format_layer_spec(input_shape, input_type)
        output['servable']['methods']['run']['output'] =\
            output.format_layer_spec(output_shape, output_type)

        output['servable']['model_summary'] = str(model)
        output['servable']['model_type'] = 'Deep NN'

        # Add torch as a dependency
        output.add_requirement('torch', torch.__version__)

        return output

    def format_layer_spec(self, layers, datatypes):
        """Make a description of a list of input or output layers

        Args:
            layers (tuple or [tuple]): Shape of the layers
            datatypes (str or [str]): Data type of each input layer
        Return:
            (dict) Description of the inputs / outputs
        """
        if isinstance(layers, tuple):
            return compose_argument_block("ndarray", "Tensor",
                                          shape=list(layers), item_type=datatypes)
        else:
            if isinstance(datatypes, str):
                datatypes = [datatypes] * len(layers)
            return compose_argument_block("tuple", "Tuple of tensors",
                                          element_types=[self.format_layer_spec(i, t)
                                                         for i, t in zip(layers, datatypes)])

    def _get_handler(self):
        return "torch.TorchServable"

    def _get_type(self):
        return "Torch Model"



"""Tools for interacting with the DLHub Search Index"""

from mdf_toolbox.globus_search.search_helper import SearchHelper
from globus_sdk.search import SearchClient
from warnings import warn


class DLHubSearchHelper(SearchHelper):
    """Helper class for building queries with DLHub"""

    def __init__(self, search_client: SearchClient, **kwargs):
        """Initialize the Helper

        Keyword arguments are passed to the underlying SearchHelper

        Args:
            search_client (SearchClient): Search client to use for authentication
        """
        super(DLHubSearchHelper, self).__init__("dlhub", search_client=search_client, **kwargs)

    def match_owner(self, owner):
        """Add a model owner to the query.

        Args:
            owner (str): The name of the owner of the model.

        Returns:
            DLHubSearchHelper: Self
        """
        if owner:
            self.match_field("dlhub.owner", owner)
        return self

    def match_servable(self, servable_name=None, owner=None, publication_date=None):
        """Add identifying model information to the query.
        If this method is called without any valid arguments, it will do nothing.

        Args:
            servable_name (str): The name of the model. **Default**: None, to match
                    all model names.
            owner (str): The name of the owner of the model. **Default**: ``None``,
                    to match all owners.
            publication_date (int): The UNIX timestamp for when the model was published.
                    **Default**: ``None``, to match all versions.

        Returns:
            DLHubSearchHelper: Self
        """
        if servable_name:
            self.match_field("dlhub.name", servable_name)
        if owner:
            self.match_owner(owner)
        if publication_date:
            self.match_field("dlhub.publication_date", publication_date)
        return self

    def match_authors(self, authors, match_all=True):
        """Add authors to the query.

        Args:
            authors (str or list of str): The authors to match.
            match_all (bool): If ``True``, will require all authors be on any results.
                    If ``False``, will only require one author to be in results.
                    **Default**: ``True``.

        Returns:
            DLHubSearchHelper: Self
        """
        if not authors:
            return self
        if isinstance(authors, str):
            authors = [authors]

        # TODO: Should we always generate creatorName when ingesting into Search or do it in SDK?
        # TODO: Potential issue: Entries without family and given name specific
        # TODO: Potential issue: Does not require first/lastname to be associated with same author
        for i, author in enumerate(authors):
            temp = author.split(",")

            # Family name is mandatory
            self.match_field(field="datacite.creators.familyName",
                             value='"{}"'.format(temp[0]), required=True if i == 0 else match_all,
                             new_group=True)

            # Given name is optional
            if len(temp) > 1:
                # If provided, it should be matched with the surname
                self.match_field(field="datacite.creators.givenName",
                                 value='"{}"'.format(temp[1].strip()), required=True,
                                 new_group=False)
        return self

    def match_domains(self, domains, match_all=True):
        """Add domains to the query.

        Args:
            domains (str or list of str): The domains to match.
            match_all (bool): If ``True``, will require all domains be on any results.
                    If ``False``, will only require one domain to be in results.
                    **Default**: ``True``.

        Returns:
            DLHubSearchHelper: Self
        """
        if not domains:
            return self
        if isinstance(domains, str):
            domains = [domains]
        # First domain should be in new group and required
        self.match_field(field="dlhub.domains", value=domains[0], required=True, new_group=True)
        # Other domains should stay in that group
        for domain in domains[1:]:
            self.match_field(field="dlhub.domains", value=domain, required=match_all,
                             new_group=False)
        return self

    def match_doi(self, doi):
        """Add a DOI to the query.

        Args:
            doi (str): The DOI to match.

        Returns:
            DLHubSearchHelper: Self
        """
        if doi:
            self.match_field("datacite.relatedIdentifiers.relatedIdentifier", '"{}"'.format(doi))
        return self


def filter_latest(results):
    """Get only the models with the most recent publication date

    Args:
        results ([dict]): List of results to filter
    Returns:
        [dict]: Only the most recent results
    """
    latest_res = {}

    # Loop over all results, get most recent publication for each servable
    for res in results:
        # TODO: Remove these warnings once search index is fixed
        if 'shorthand_name' not in res['dlhub']:
            warn('Found entries in DLHub index that lack shorthand_name. '
                 'Please contact DLHub team', RuntimeWarning)
            continue
        if 'publication_date' not in res['dlhub']:
            warn('Found entries in DLHub index that lack publication_date.'
                 ' Please contact DLHub team', RuntimeWarning)
            continue
        ident = res["dlhub"]["shorthand_name"]
        pub_date = int(res["dlhub"]["publication_date"])

        # If res not in latest_res, or res version is newer than latest_res
        # TODO: Make publication_date an integer in Search
        if latest_res.get(ident, ({}, -1))[1] < pub_date:
            latest_res[ident] = (res, pub_date)

    # Return only the most recent models
    return [r[0] for r in latest_res.values()]


def get_method_details(metadata, method_name=None):
    """Get the method details for use by humans

    Gets only the method fields out of the metadata record for an objecvt,
    removes the "method_details" field, which is used only during construction
    of the object.

    Will either return the data for all methods or, if ``method_name`` is provided,
    only a single function

    Args:
        metadata (dict): Metadata record for a servable
        method_name (str): Optional: Name of the function to retrieve
    Returns:
        dict: Metadata for the servable method, or a single method of ``method_name`` is used
    """

    # Get the "methods" block
    methods = metadata['servable']['methods']

    # Remove "method_details"
    for m in methods.values():
        m.pop('method_details', None)

    # If desired, return only a single method
    if method_name is not None:
        if method_name not in methods:
            raise ValueError('No such method: {}'.format(method_name))
        return methods[method_name]
    return methods



"""Tools for dealing with asynchronous execution"""
from globus_sdk import GlobusAPIError
from concurrent.futures import Future
from threading import Thread
from time import sleep


class DLHubFuture(Future):
    """Utility class for simplifying asynchronous execution in DLHub"""

    def __init__(self, client, task_id: str, ping_interval: float):
        """
        Args:
             client (DLHubClient): Already-initialized client, used to check
             task_id (str): Set the task ID of the
             ping_interval (float): How often to ping the server to check status in seconds
        """
        super().__init__()
        self.client = client
        self.task_id = task_id
        self.ping_interval = ping_interval

        # List of pending statuses returned by funcX.
        # TODO: Replace this once funcX stops raising exceptions when a task is pending.
        self.pending_statuses = ["received", "waiting-for-ep", "waiting-for-nodes",
                                 "waiting-for-launch", "running"]

        # Once you create this, the task has already started
        self.set_running_or_notify_cancel()

        # Forcing the ping interval to be no less than 1s
        if ping_interval < 1:
            assert AttributeError('Ping interval must be at least 1 second')

        # Start a thread that polls status
        self._checker_thread = Thread(target=DLHubFuture._ping_server, args=(self,))
        self._checker_thread.start()

    def _ping_server(self):
        while True:
            sleep(self.ping_interval)
            try:
                if not self.running():
                    break
            except GlobusAPIError:
                # Keep pinging even if the results fail
                continue

    def running(self):
        if super().running():
            # If the task isn't already completed, check if it is still running
            try:
                status = self.client.get_result(self.task_id, verbose=True)
            except Exception as e:
                # Check if it is "Task pending". funcX throws an exception on pending.
                if e.args[0] in self.pending_statuses:
                    return True
                else:
                    self.set_exception(e)
                    return False

            if isinstance(status, tuple):
                # TODO pass in verbose setting?
                self.set_result(status[0])
                return False

        return False

    def stop(self):
        """Stop the execution of the function"""
        # TODO (lw): Should be attempt to cancel the execution of the task on DLHub?
        self.set_exception(Exception('Cancelled by user'))



from dlhub_sdk.models import BaseMetadataModel


class BaseServableModel(BaseMetadataModel):
    """Base class for servables"""

    def __init__(self):
        super(BaseServableModel, self).__init__()

        # Add the resource type
        #  I chose "InteractiveResource" as the point of DLHub is to provide
        #   web servies for interacting with these models ("query/response portals" are
        #   defined as "InteractiveResources") rather than downloading the source code
        #   (which would fit the definition of software)
        self._output['datacite']['resourceType'] = {'resourceTypeGeneral': 'InteractiveResource'}

        # Define artifact type
        self._output['dlhub']['type'] = 'servable'

        # Initialize the model running-information
        self._output['servable'] = {
            'methods': {'run': {}},
            'shim': self._get_handler(),
            'type': self._get_type()
        }

    def _get_handler(self):
        """Generate the name of the servable class that DLHub will use to read this metadata

        Returns:
            (string) path to the python class (e.g., "python.PythonStaticMethod")
        """
        raise NotImplementedError()

    def _get_type(self):
        """Get a human-friendly name for this type of servable

        Returns:
            (string): Human-friendly name of an object
        """
        raise NotImplementedError()

    def register_function(self, name, inputs, outputs, parameters=None, method_details=None):
        """Registers a new function to this servable

        See :code:`compose_argument_type` utility function for how to define the inputs
        and outputs to this function.

        Args:
            name (string): Name of the function (e.g., "run")
            inputs (dict): Description of inputs to the function
            outputs (dict): Description of the outputs of the function
            parameters (dict): Any additional parameters for the function and their default values
            method_details (dict): Any options used when constructing a shim to run this function.
        """

        # Check defaults
        if method_details is None:
            method_details = {}
        if parameters is None:
            parameters = {}

        # Add the function
        self._output["servable"]["methods"][name] = {
            'input': inputs,
            'output': outputs,
            'parameters': parameters,
            'method_details': method_details
        }

        return self

    def to_dict(self, simplify_paths=False, save_class_data=False):
        # Make sure the inputs and outputs have been set
        if len(self._output["servable"]["methods"]["run"].get("input", {})) == 0:
            raise ValueError('Inputs have not been defined')
        if len(self._output["servable"]["methods"]["run"].get("output", {})) == 0:
            raise ValueError('Outputs have not been defined')

        return super(BaseServableModel, self).to_dict(simplify_paths)



try:
    # Attempt to use regular keras first, as we figure it's installed for a reason
    import keras
except ImportError:
    from tensorflow import keras


from dlhub_sdk.models.servables.python import BasePythonServableModel
from dlhub_sdk.utils.types import compose_argument_block


keras_version = keras.__version__
_keras_version_tuple = tuple(int(i) for i in keras_version.rstrip("-tf").split("."))
_summary_limit = 10000


def _detect_backend(output):
    """Add the backend

    Args:
        output (KerasModel): Current description of Keras model, will be modified
    """

    # Determine the name of the object
    my_backend = keras.backend.backend().lower()

    # Add it as a requirement
    output.add_requirement(my_backend, 'detect')


class KerasModel(BasePythonServableModel):
    """Servable based on a Keras Model object.

    Assumes that the model has been saved to an hdf5 file"""

    @classmethod
    def create_model(cls, model_path, output_names=None, arch_path=None,
                     custom_objects=None):
        """Initialize a Keras model.

        Args:
            model_path (string): Path to the hd5 file that contains the weights and, optionally,
                the architecture
            output_names ([string] or [[string]]): Names of output classes.
            arch_path (string): Path to the hd5 model containing the architecture, if not
                available in the file at :code:`model_path`.
            custom_objects (dict): Map of layer names to custom layers. See
                `Keras Documentation
                <https://www.tensorflow.org/api_docs/python/tf/keras/models/load_model>`_
                for more details.
       """
        output = super(KerasModel, cls).create_model('predict')

        # Add model as a file to be sent
        output.add_file(model_path, 'model')
        if arch_path is not None:
            output.add_file(arch_path, 'arch')

        # Store the list of custom objects
        if custom_objects is not None:
            for k, v in custom_objects.items():
                output.add_custom_object(k, v)

        # Get the model details
        if arch_path is None:
            model = keras.models.load_model(model_path, custom_objects=custom_objects)
        else:
            if arch_path.endswith('.h5') or arch_path.endswith('.hdf') \
                    or arch_path.endswith('.hdf5') or arch_path.endswith('.hd5'):
                model = keras.models.load_model(arch_path,
                                                custom_objects=custom_objects, compile=False)
            elif arch_path.endswith('.json'):
                with open(arch_path) as fp:
                    json_string = fp.read()
                model = keras.models.model_from_json(json_string, custom_objects=custom_objects)
            elif arch_path.endswith('.yml') or arch_path.endswith('.yaml'):
                with open(arch_path) as fp:
                    yaml_string = fp.read()
                model = keras.models.model_from_yaml(yaml_string, custom_objects=custom_objects)
            else:
                raise ValueError('File type for architecture not recognized')
            model.load_weights(model_path)

        # Get the inputs of the model
        output['servable']['methods']['run']['input'] = output.format_layer_spec(model.input_shape)
        output['servable']['methods']['run']['output'] = output.format_layer_spec(
            model.output_shape)
        if output_names is not None:
            output['servable']['methods']['run']['method_details']['classes'] = output_names

        # Get a full description of the model. Limit summary to _summary_limit in length
        output.summary = ""

        def capture_summary(x):
            output.summary += x + "\n"

        model.summary(print_fn=capture_summary)
        output.summary = (output.summary[:_summary_limit] + '<<TRUNCATED>>') \
            if len(output.summary) > _summary_limit else output.summary

        output['servable']['model_summary'] = output.summary
        output['servable']['model_type'] = 'Deep NN'

        # Add keras as a dependency
        if not keras_version.endswith("-tf"):
            output.add_requirement('keras', keras_version)
        output.add_requirement('h5py', 'detect')

        # Detect backend and get its version
        _detect_backend(output)

        return output

    def format_layer_spec(self, layers):
        """Make a description of a list of input or output layers

        Args:
            layers (tuple or [tuple]): Shape of the layers
        Return:
            (dict) Description of the inputs / outputs
        """
        if isinstance(layers, tuple):
            return compose_argument_block("ndarray", "Tensor", shape=list(layers))
        else:
            return compose_argument_block("tuple", "Tuple of tensors",
                                          element_types=[self.format_layer_spec(i) for i in layers])

    def add_custom_object(self, name, custom_object):
        """Add a custom layer to the model specification

        See `Keras FAQs
        <https://keras.io/getting-started/faq/#handling-custom-layers-or-other-custom-objects-in-saved-models>`
        for details.

        Args:
              name (string): Name of the custom object
              custom_object (class): Class of the custom object
        Return:
            self
        """

        # get the class name for the custom object
        object_name = custom_object.__name__
        module = custom_object.__module__

        # Add the layer to the model definition
        if 'options' not in self._output['servable']:
            self['servable']['options'] = {}
        if 'custom_objects' not in self['servable']['options']:
            self['servable']['options']['custom_objects'] = {}
        self['servable']['options']['custom_objects'][name] = '{}.{}'.format(module, object_name)

    def _get_handler(self):
        return "keras.KerasServable"

    def _get_type(self):
        return "Keras Model"
